class State:
    def __init__(self):
        self.buffer = ''

from .link import LinkText
from .link import LinkUrl
from .inlinecode import InlineCode
from .text import Text
from .title import Title
from .image import Image
from .tick import Tick
from .codeblock import CodeBlock
from .link import Link


import sys
from .tokenizer import Tokenizer
from .parser import Parser

def process_file(file_path):
    tokens = None
    tokenizer = Tokenizer()

    with open(file_path, 'r') as f:
        tokens = tokenizer.scan(f) # todo: add timeout

    return Parser().parse(tokens)

if __name__ == '__main__':
    print(process_file(sys.argv[1]))













from . import State


class InlineCode(State):
    def __init__(self):
        self.name = 'InlineCode'
        super().__init__()

    def parse(self, stack):
        tik = stack.pop() # drop first tick
        if tik.value != '`':
            raise Exception('Inline code must start with a tick')
        if stack.peek_type() == 'Tick':
            raise Exception('Inline code must not be empty')
        # We seek until the next tick. 
        # This will terminate the inline code no matter the number of ticks
        # eg. `this is inline code``` is valid and closed
        tokens = stack.seek_until(lambda tok: tok.type == 'Tick')
        l = len(tokens)
        tokens = tokens[:l-1] # remove the final tick

        self.buffer += '<code class="inline_code">'
        self.buffer += ''.join([t.value for t in tokens])
        self.buffer += '</code>'
        return self.buffer


from . import State
from .title import Title
from .tick import Tick


class Plain(State):
    '''
    This is the root state in the AST.
    '''
    def __init__(self):
        super().__init__()
        self.name = 'Plain'
    
    def parse(self, stack):
        while not stack.is_empty():
            token = stack.peek()

            # todo: this will choke when token in middle of text
            # eg. "this is not # a title"
            match token.type: 
                case 'Octo': self.buffer += Title().parse(stack)
                case 'Tick': self.buffer += Tick().parse(stack)
                case _: self.buffer += stack.pop().value
        return self.buffer


from .structures import Token

class Tokenizer:
    def __init__(self):
        self.tokens: list[Token] = []
        self.digits = [n for n in '0123456789']

    def append(self, type, chr):
        '''Append a token to the token list.
        If the type of the token to append is the same as the last token, 
        append the value to the last token instead of the list.
        '''
        if len(self.tokens) > 0:
            if self.tokens[-1].type == type: # if previous token is same type, append to value
                self.tokens[-1].value += chr
                return

        self.tokens.append(Token(type, chr))

    def scan(self, reader):
        while True:
            c = reader.read(1)
            if c == '': # EOF
                return self.tokens
            if c in self.digits: # number
                self.append('Digit',c)
                continue # no idea how to do this in the match statement
            match c:
                case '!': self.append('Bang', c)
                case '#': self.append('Octo',c)
                case ' ': self.append('Space',c)
                case '\n': self.append('Newline',c)
                case '`': self.append('Tick',c)
                case '[': self.append('OpenBracket',c)
                case ']': self.append('CloseBracket',c)
                case '(': self.append('OpenParen',c)
                case ')': self.append('CloseParen',c)
                case '\\': self.append('Escape',c)
                case _: self.append('Text',c)



from . import State
import string
import re
from urllib.request import urlopen

class CodeBlock(State):
    def __init__(self):
        self.name = 'CodeBlock'
        super().__init__()
        self.valid_language_chars = list(string.ascii_lowercase)
        self.direct_pattern = r'^```[a-z]+\n$'

    def _is_direct(self, stack):
        tokens = stack.peek(n=3)
        opening = ''.join([t.value for t in tokens])
        return re.match(self.direct_pattern, opening)

    def _parse_direct(self, stack):
        stack.pop() # drop opening ```
        language = stack.pop().value
        # todo: a second language specific parse to break up the code into elements
        tokens = stack.seek_until(lambda tok: tok.value.startswith('```'), include_last=False)
        stack.pop() # drop closing ```
        inner = ''.join([t.value for t in tokens])
        return self.finalize(inner, language=language)

    def finalize(self, inner, language=None):
        self.buffer += f"<pre class=\"code_block\"><code class=\"{language}\">"
        self.buffer += inner
        self.buffer += '</code></pre>'
        return self.buffer

    def parse(self, stack):
        if self._is_direct(stack):
            return self._parse_direct(stack)

        #todo: support gist links
        raise Exception(
            'Code block must start with ``` and specify language ending in newline.')


from . import State


class Link(State):
    def __init__(self):
        super().__init__()
        self.name = 'Link'

    def parse(self, stack):
        text = LinkText().parse(stack)
        link = LinkUrl().parse(stack)
        return f"<a href='{link}'>{text}</a>"


class LinkText(State):
    def __init__(self):
        super().__init__()
        self.name = 'LinkText'

    def parse(self, stack):
        stack.pop() # drop open bracket
        tokens = stack.seek_until(lambda tok: tok.type == 'CloseBracket')
        tokens.pop() # drop close bracket
        return ''.join([t.value for t in tokens])


class LinkUrl(State):
    def __init__(self):
        super().__init__()
        self.name = 'LinkUrl'

    def parse(self, stack):
        stack.pop() # drop open parenthesis
        tokens = stack.seek_until(lambda tok: tok.type == 'CloseParen')
        tokens.pop() # drop close bracket
        return ''.join([t.value for t in tokens])


from .structures import Stack
from .states import Title, Tick, Text
class Parser:
    '''
    BNF
    <Plain> ::= Text | Title | CodeBlock | InlineCode
    <Text> ::= InlineCode | Link | Tick | Text
    <Title> ::= Octo, Space, Text, Newline
    <CodeBlock> ::= Tick, Newline, ANY+, Newline, Tick, Newline
    '''
    def parse(self, tokens):
        stack = Stack(tokens)
        if stack.stack[0].type != 'Newline':
            raise Exception("Tark files must end in a newline")
        document = []
        try:
            while not stack.is_empty():
                token = stack.peek()
                match token.type:
                    case 'Octo': document.append(Title().parse(stack))
                    case 'Tick': document.append(Tick().parse(stack))
                    case 'Text': document.append("<div>"+Text().parse(stack)+"</div>")
                    case 'Newline': document.append(stack.pop().value) # out of context newline
                    case 'Space': document.append(stack.pop().value) # out of context space
                    case 'Digit': document.append(stack.pop().value) # out of context digit
                    case _: raise Exception('Unexpected token type {}'.format(token.type))
            return ''.join(document)
        except Exception as e:
            raise Exception(f"Parse error at [{stack.line},{stack.row}]\n", e)


from . import State
from .inlinecode import InlineCode
from .link import Link
from .image import Image
class Text(State):
    def __init__(self):
        super().__init__()
        self.name = 'Text'
    def parse(self, stack):
        exit = False
        while not stack.is_empty():
            token = stack.peek()
            match token.type:
                case 'OpenBracket': self.buffer += Link().parse(stack)
                case 'Tick': self.buffer += InlineCode().parse(stack)
                case 'Bang': self.buffer += Image().parse(stack)
                case 'Escape': self.buffer += stack.pop(n=2)[1].value # append next as literal
                case 'Newline':
                    stack.pop() # drop and return
                    return self.buffer
                case _: 
                    self.buffer += stack.pop().value
        return self.buffer



class Token:
    def __init__(self, type: str, value: str, line=0, row=0):
        self.type = type
        self.value = value
        self.line = line
        self.row = row
        self.length = len(value)

    def __repr__(self):
        return f"Token({self.type}, {repr(self.value)})"

    def __eq__(self, other): # type compare only
        return self.type == other.type

class Stack:
    '''
    Internally a list. This struct provides 'stack' style behavior 
    and some custom functions.
    '''
    stack = []
    row = 0
    line = 0

    def __init__(self, tokens: list[Token]):
        tokens.reverse()
        self.stack = tokens

    def size(self):
        '''Return the size of the stack.'''
        return len(self.stack)
    def push(self, item: Token):
        '''Put item on the stack.'''
        self.stack.append(item)

    def peek(self, n=None):
        '''
        Peek 1 or N tokens from the stack.
        Raise exception if stack is empty.
        '''
        if self.is_empty():
            raise Exception("Attempted to peek at empty stack.")
        if not n:
            return self.stack[-1]
        
        buf = []
        for i in range(1,n+1,1):
            buf.append(self.stack[-i])
        return buf

    def peek_type(self, n=1):
        '''Peek 1 or N of the next token types in the stack.'''
        tokens = self.peek(n)

        for i in range(n):
            tokens[i] = tokens[i].type
        return tokens

    def is_empty(self):
        return len(self.stack) == 0
    def __str__(self):
        return str(self.stack)
    
    def pop(self, n=1):
        buf = []
        for i in range(n):
            self.row += 1
            token = self.stack.pop()
            if token.type == 'Newline':
                self.line += 1
                self.row = 0
            else:
                self.row += token.length
            buf.append(token)

        if n==1:
            return buf[0]
        return buf
    
    def seek_until(self, func, include_last=True, error_on_empty=True)-> list[Token]:
        '''
        From the HEAD of stack, pop until provided lambda(token) is true.
        Return the popped tokens.
        '''
        buffer = []
        while True:
            if self.is_empty(): # exit condition 1
                if error_on_empty:
                    raise Exception('Stack empty before seek_until found match.')
                else:
                    return buffer
            token = self.pop()
            buffer.append(token)
            if func(token): # exit condition 2
                if not include_last:
                    self.push(buffer.pop()) # put back the last token
                return buffer


from . import State
from .link import LinkUrl
from .link import LinkText

class Image(State):
    def __init__(self):
        self.name = 'Image'
        super().__init__()

    def parse(self, stack):
        stack.pop() # drop the bang
        text = LinkText().parse(stack)
        link = LinkUrl().parse(stack)
        return f"<img src='{link}' alt='{text}'>"


from . import State


class Title(State):
    def __init__(self):
        self.name = 'Octo'
        super().__init__()

    def parse(self, stack):
        # first tokens are Octo+, Space
        token = stack.pop()
        size = len(token.value)
        if size > 6:
            raise Exception('Octo title cannot be greater than size 6')
        if '#'*size != token.value:
            raise Exception('Octo title must be a series of #')

        token = stack.pop() # dump the space
        if token.type != 'Space':
            raise Exception('Octo title must be followed by a space')
        
        tokens = stack.seek_until(lambda tok: tok.value.startswith('\n\n'))
        l = len(tokens)
        tokens = tokens[:l-1] # remove the final newline

        self.buffer += f"<h{size}>"
        self.buffer += ''.join([t.value for t in tokens])
        self.buffer += f"</h{size}>"
        return self.buffer


from setuptools import setup, find_packages

with open("README.md", "r") as readme_file:
    readme = readme_file.read()

setup(
    name="tark",
    version="0.0.4",
    author="tyler jones",
    description="Markdown parser focused on blogs.",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/tylerhjones/tark/",
    packages=find_packages(),
    install_requires=[],
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
    ],
)


from . import State
from .inlinecode import InlineCode
from .codeblock import CodeBlock

class Tick(State):
    def __init__(self):
        self.name = 'Tick'
        super().__init__()

    def parse(self, stack):
        token = stack.peek()

        match token.value:
            case '`':
                return InlineCode().parse(stack)
            case '```':
                return CodeBlock().parse(stack)
            case _:
                raise Exception('Tick length must be 1 or 3')